#ifndef IRRAM_REALS_H
#define IRRAM_REALS_H

#ifdef LR
#include "LR_interface.h"
#else 
#ifdef LRGMP
#include "LRGMP_interface.h"
#else 
#ifdef GMP
#include "GMP_interface.h"
#else 
#ifdef MPFR
#include "MPFR_interface.h"
#else
#include "Backend not defined"
#endif
#endif
#endif
#endif


extern "C" void iRRAM_initialize(int argc,char** argv);
extern void iRRAM_exec(void iRRAM());



// Definition of class DYADIC ******************

class DYADIC
{
public:

// Set absolute precision for operations:  -----

static void setprec(long p) {precision=p;};

// Constructors: -------------------------------

DYADIC ();
DYADIC (const long i);
DYADIC (const DYADIC& y);

// Copy Constructor: ---------------------------


DYADIC&   operator = (const DYADIC& y);

// Destructor: ---------------------------------

~DYADIC();

// Standard Arithmetic: ------------------------

friend DYADIC  operator +  (const DYADIC& x,
                            const DYADIC& y);

friend DYADIC  operator -  (const DYADIC& x,
                            const DYADIC& y);

friend DYADIC  operator -  (const DYADIC& x);

friend DYADIC  operator *  (const DYADIC& x,
                            const DYADIC& y);

friend DYADIC  operator /  (const DYADIC& x,
                            const DYADIC& y);

// Comparisons: --------------------------------

friend bool    operator <  (const DYADIC& x,
                            const DYADIC& y);

friend bool    operator <= (const DYADIC& x,
                            const DYADIC& y);

friend bool    operator >  (const DYADIC& x,
                            const DYADIC& y);

friend bool    operator >= (const DYADIC& x,
                            const DYADIC& y);

friend bool    operator == (const DYADIC& x,
                            const DYADIC& y);

friend bool    operator != (const DYADIC& x,
                            const DYADIC& y);

// Output: -------------------------------------

friend void    rwrite      (const DYADIC& x,
                            const long w);

// miscellaneous: ------------------------------

friend long    size        (const DYADIC& x); 

friend DYADIC  abs         (const DYADIC& x); 

// coexistence with other classes: -------------

friend class REAL;

friend DYADIC  approx      (const REAL& x, 
                            const long p);

// implementational issues: --------------------

MP_type    value; 


private:

static long precision;


DYADIC(MP_type  y);

};

// Definition of class METRIC_OBJECT ********************
#define SIZETYPEMANTISSA unsigned long
#define SIZETYPEEXPONENT long

struct sizetype { SIZETYPEMANTISSA  mantissa; SIZETYPEEXPONENT exponent; } ;

class METRIC_OBJECT {
public:
METRIC_OBJECT();
virtual ~METRIC_OBJECT();
virtual void clear(){};
virtual void adderror (sizetype error){};
virtual void seterror (sizetype error){};
virtual void geterror (sizetype& error) const{};
METRIC_OBJECT* next;
METRIC_OBJECT* prev;
};


// Definition of class REAL ********************
class REAL : public  METRIC_OBJECT
{
public:

// Constructors: -------------------------------

REAL();
REAL(int  i);
REAL(long i);
REAL(char* s);
REAL(const DYADIC& y);
REAL(const REAL& y);
REAL(double d);

// Copy Constructor: ---------------------------

REAL& operator = (const REAL& y);

// Destructor: ---------------------------------

~REAL();
void clear();

// Standard Arithmetic: ------------------------

friend REAL  operator +  (const REAL& x,
                          const REAL& y);
friend REAL& operator += (REAL& x,
                          const REAL& y);
			  
friend REAL  operator -  (const REAL& x,
                          const REAL& y);
friend REAL  operator -  (const REAL& x);

friend REAL  operator *  (const REAL& x,
                          const REAL& y);
friend REAL  operator *  (const REAL& x,
                          long int n);
friend REAL  operator *  (const REAL& x,
                          int n);
			  
friend REAL& operator *= (REAL& x,
                          const REAL& y);
friend REAL& operator *= (REAL& x,
                          long int n);
friend REAL& operator *= (REAL& x,
                          int n);
			  
friend REAL  operator /  (const REAL& x,
                          const REAL& y);
friend REAL  operator /  (const REAL& x,
                          long int n);
friend REAL  operator /  (const REAL& x,
                          int n);

friend REAL  operator ^  (const REAL& x,
                          const REAL& y);

friend REAL  operator ^  (const REAL& x,
                          long n);

friend REAL  operator ^  (const REAL& x,
                          int n);

friend REAL  operator << (const REAL& x,
                          long n);

friend REAL  operator >> (const REAL& x,
                          long n);

friend REAL  sqrt        (const REAL& x);

friend REAL  square      (const REAL& x);

friend REAL  scale       (const REAL& x,
                          long k);

// Comparisons: --------------------------------


friend bool  operator <  (const REAL& x,
                          const REAL& y);
friend bool  operator <= (const REAL& x,

                          const REAL& y);
friend bool  operator >  (const REAL& x,
                          const REAL& y);
friend bool  operator >= (const REAL& x,
                          const REAL& y);

friend bool  positive    (const REAL& x,
                          const long k);

friend bool  bound       (const REAL& x,
                          const long k);

friend DYADIC approx     (const REAL& x, 
                          const long p);

// Output: -------------------------------------

friend void   rwrite     (const REAL& x,
                          const long p,
                          const long w);

friend void   rwrite     (const REAL& x,
                          const long w);

friend void   rshow      (const REAL& x,
                          const long w);

void REAL::rcheck() const;

friend long    upperbound (const REAL& x);
friend long    size       (const REAL& x); 



friend REAL    round2     (const REAL& x); 
friend long    round      (const REAL& x); 

friend REAL    abs        (const REAL& x);
 

// limit operators: ------------------------

//friend REAL    limit      (REAL f(long, const REAL&),
//                           const REAL& x);

friend REAL    limit      (REAL f(long, const REAL&, const REAL&),
                           const REAL& x, 
                           const REAL& y);

friend REAL    limit      (REAL f(long));

friend REAL    limit_hint (REAL f(long, const REAL&),
                           long hint,
                           const REAL& x);

friend REAL    limit_hint (REAL f(long, const REAL&, const REAL&),
                           long hint,
                           const REAL& x, 
                           const REAL& y);

friend REAL    limit_hint (REAL f(long),
                           long hint);

friend REAL    limit_lip  (REAL f(long, const REAL&),
                           long lip,
                           const REAL& x);

friend REAL    limit_lip   (REAL f(long, const REAL&, const REAL&),
                           long lip,
                           const REAL& x,
                           const REAL& y);

friend REAL    iterate     (REAL (*)(REAL&, REAL&, const REAL&),
                           const REAL&, const REAL&, const REAL&);

friend REAL    iterate     (REAL (*)(REAL&, REAL&),
                           const REAL&, const REAL&);

// reduced error propagation: ------------------------

friend REAL    lipschitz  (REAL f(const REAL&),
                           long lip,
                           const REAL& x);

friend REAL    lipschitz  (REAL f(const REAL&,const REAL&),
                           long lip,
                           const REAL& x,
                           const REAL& y);

friend REAL    lipschitz  (REAL f(long, const REAL&),
                           long lip,
                           long k,
                           const REAL& x);

friend REAL    lipschitz  (REAL f(long, const REAL&,const REAL&),
                           long lip,
                           long k,
                           const REAL& x,
                           const REAL& y);

// coexistence with other classes: -------------

friend class REALMATRIX;
friend class SPARSEREALMATRIX;

// implementational issues: --------------------

private:

MP_type       value; 
sizetype      error;

REAL(MP_type y, const sizetype errorinfo);

public:
void adderror (sizetype error);
void seterror (sizetype error);
void geterror (sizetype& error) const;

};


void rscanf  (const char *format, void * input);
void rprintf (const char *rformat, ...);

void continous_begin ();
void continous_end   ();


#define ABSOLUTE 0
#define RELATIVE 1
void precision_policy (long policy);

void stiff_begin ();
void stiff_end   ();




#endif /* ! IRRAM_REALS_H */
