#include "stdio.h"
#include "stdlib.h"
#include "REALLIB.h"
#include "REALlimits.h"
#include "REALcore.h"

REAL exp_approx (long prec, const REAL& x){
  REAL xs=x/ln2();
  REAL z;

  long s=round(xs)+1;
  long it,wd;

  long min_bound=prec-s;
  if (min_bound > -2) min_bound=-2;
  
  wd= 1+ long(sqrt(double(-prec)))/16;
  
  if ( wd <2 )    it=long(log(double(-prec)))*2;
  else            it=long(log(double(-prec)))*4;


  xs=scale(x-ln2()*s,-it);

  if ( wd > 1 ) {
    REAL xpow[wd],xf=power(xs,wd);
    xpow[0]=1;
    z=1;
    for (long i=1; i<wd; i++) {
      xpow[i]=xpow[i-1]*xs/i;
      z+=xpow[i];
      }
    REAL xp=1,f=1,e=z;
    long i=1;
    long t=upperbound(xf),t1=0;
    while ( !bound(e,min_bound) ) {
      t1=t1+t;
      xp*=xf; 
      e=0;
      for (long j=0; j<wd; j++) {
        f=1;
        for (long k=i;k<i+wd;k++) f*=k;
        xpow[j]=scale(xpow[j],t)/f;
        i=i+1;
        e+=xpow[j];
        }
      e*=scale(xp,-t1);
      z+=e;
      }
    }
  else {
    REAL e=xs;
    z=1+e;
    long i=1;
    while ( !bound(e,min_bound) ) {
      i+=1;
      e*=(xs/i); 
      z+=e;
      }
    }

  for (long i=1; i<= it;i+=1)z*=z;

  return scale(z,s);
};

REAL exp (const REAL& x)
{
  if ( positive(x-750001,0)  ) {fprintf(stderr,"Overflow in exp(x)\n");exit(1);} 
  continous_begin ();
  long i=(long)((round(x)+1)*1.443);
  REAL y= limit_lip(exp_approx,i,x);
  continous_end ();
  return y;
};

REAL euler_approx (long prec)
{
  if ( prec >= 2 ) return 0;
  REAL z=1,y=z;
  long i=1;
  while ( !bound(y,prec-1) ) {
    y=y/i;
    z=z+y;
    i+=1;
  }
  return z;
};

REAL euler ()
{
  return limit(euler_approx);
};



REAL log_taylor_approx(long prec, const REAL& x0) {
// Computing series expansion for log(x) for x > 1/2
   REAL x=x0;
//  printf("####log ser#\n");
   long it=(long)(sqrt((double) -prec/4)); 
   for (long i=1;i<=it;i+=1) x=sqrt(x);
   REAL xm1=x-1;
   REAL y=xm1/x;
   REAL y0=y*y,y1=y*y0,y2=y0*y0,y3=y*y2,y4=y1*y1,yf=y3,yp=1;
   REAL s=y;
   long i=7;
  REAL z=y+y0/2+y1/3+y2/4+y3/5+y4/6;
   while ( !bound(s,prec-it) ) {
    yp*=yf;
    s=yp*(y0/i+y1/(i+1)+y2/(i+2)+y3/(i+3)+y4/(i+4));
    i=i+5;
    z+=s;
   }
   return scale(z,it);
;
}


REAL iterate
   (REAL f(REAL&, REAL&),
   REAL const& a_0, REAL const& b_0)
{ REAL a=a_0,b=b_0,e;
sizetype a_error,e_error;
continous_begin();
do {
  e=f(a,b);
} while (! bound(e,15*actual_prec/16));
a.geterror(a_error);
e.geterror(e_error);
sizetype_inc(a_error,e_error);
MP_getsize(e.value,e_error);
sizetype_inc(a_error,e_error);
a.seterror(a_error); 
continous_end();
return a;
}

REAL agm(REAL& a,REAL& b){
  REAL e=a*b;
  a=scale(a+b,-1); 
  b=sqrt(e);
  return a-b;
}



REAL log_agm_approx(long prec, const REAL& z) {
  long N=25-prec/2;
//  printf("####log AGM#\n");
  precision_policy(7);
  REAL delta1= scale(z,-N);
  REAL l = ln2()*(N+2)-scale(pi()/iterate(agm,REAL(1),delta1),-1);
  precision_policy(0);
  return l;
}

REAL log_approx(long prec, const REAL& z) {
  if ( prec > -100) return log_taylor_approx(prec,z);
  return log_agm_approx(prec,z);
}


REAL log(const REAL& x) {
 long s=size(x);
 REAL y= sqrt(sqrt(scale(x,2-s)));
 y=limit_lip(log_approx,2,y);
 return 4*y+(s-2)*ln2();
}


